from typing import Union, Dict, List

from loguru import logger

from planqk.commons.runtime.json import json_to_dict


class FileReader:
    def __init__(self, file_paths: List[str], base64_encoded: bool = False):
        self.__file_paths = file_paths
        self.__base64_encoded = base64_encoded

        self.__file_path = None
        self.__file_object = None

    def __enter__(self):
        # open first file in list that exists
        for file_path in self.__file_paths:
            try:
                self.__file_path = file_path
                self.__file_object = open(file_path, "r")
                return self
            except FileNotFoundError:
                continue

        raise FileNotFoundError(f"No file found in paths: {self.__file_paths}")

    def __exit__(self, resource_type, value, tb):
        self.__file_object.close()

    def file(self):
        if self.__file_path is None or self.__file_object is None or self.__file_object.closed:
            raise RuntimeError("FileReader must be used with the 'with' keyword")
        return self.__file_object

    def read_to_string(self) -> Union[str, None]:
        try:
            return self.file().read()
        except Exception as e:
            logger.error("Error reading file '{}': {}", self.__file_path, e)
            return None

    def read_to_dict(self) -> Union[Dict, None]:
        data = self.read_to_string()
        try:
            return json_to_dict(data, self.__base64_encoded)
        except ValueError as e:
            logger.error("Error reading file '{}': {}", self.__file_path, e)
            return None



from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple, Union


def _find_user_code(entry_point: Union[str, Path]) -> Tuple[Path, Optional[str], Optional[str]]:
    """
    Generate a (path, package, method) triple from the given entry point.

    Args:
        entry_point (Union[str, Path]): the entry point to execute

    Raises:
        ValueError: If the entry point could not be parsed

    Returns:
        Tuple[Path, Optional[str], Optional[str]]: the (path, package, method) triple
    """
    if isinstance(entry_point, Path):
        if not entry_point.is_file():
            raise ValueError("If entry pint is a Path object it must point to a file!")
        path = entry_point.parent
        package = entry_point.name
        if package.endswith(".py"):
            package = package[:-3]
        return path, package, None

    *path, rest = entry_point.replace("\\", "/").split("/")
    if not rest:
        raise ValueError("Cannot parse entry point '{}'".format(entry_point))
    if rest.count(':') > 1:
        raise ValueError("Cannot parse entry point '{}'".format(entry_point))

    package, *method_list = rest.split(":", maxsplit=1)

    method = method_list[0] if method_list else None

    if package.endswith('.py'):
        package = package[:-3]
    return Path('/'.join(path)), package, method


def _run_package_as_main(package: str):
    """
    Run the given package as __main__.

    Args:
        package (str): the package to import and run
    """
    import runpy
    runpy.run_module(package, init_globals={}, run_name="__main__")


def _run_method_in_package(
        package: str, method: str,
        entry_point_arguments: Optional[Dict[str, Union[Sequence[Any], Dict[str, Any]]]] = None
) -> Any:
    """
    Run a specific method of a package.

    Args:
        package (str): The package to import
        method (str): the method to run in that package
        entry_point_arguments ({"args": Sequence[Any], "kwargs": Dict[str, Any]}): A dict containing the positional and keyword arguments to pass to the entry point function. Defaults to None.
    """
    from importlib import import_module

    imported_package = import_module(package)

    run_method = getattr(imported_package, method)
    if entry_point_arguments is None:
        return run_method()

    # run method with arguments
    return run_method(*entry_point_arguments.get("args", []), **entry_point_arguments.get("kwargs", {}))


def _run_user_code(
        entry_point: Union[str, Path],
        entry_point_arguments: Optional[Dict[str, Union[Sequence[Any], Dict[str, Any]]]] = None
) -> Optional[Any]:
    """
    Run user code from the given entry_point.

    Args:
        entry_point (Union[str, Path]): the entry point of the code to run
        entry_point_arguments ({"args": Sequence[Any], "kwargs": Dict[str, Any]}): A dict containing the positional and keyword arguments to pass to the entry point function. Defaults to None.

    Raises:
        ValueError: If the entry point could not be parsed or found on disk
    """
    path, package, method = _find_user_code(entry_point=entry_point)

    if not path.exists() or not path.is_dir():
        raise ValueError("Path {} of entry point {} not found!".format(path, entry_point))

    if not package:
        raise ValueError("Cannot run empty package.")

    if path != Path("interceptor_framework"):
        # update python path to make imports of the user code succeed
        import sys
        sys.path.insert(0, str(path.absolute()))

    if not method:
        _run_package_as_main(package)
        return None
    else:
        return _run_method_in_package(package, method=method, entry_point_arguments=entry_point_arguments)


def run(entry_point: str, data: Optional[Dict], params: Optional[Dict]) -> any:
    if data is None:
        data = dict()
    if params is None:
        params = dict()
    entry_point_arguments = {"kwargs": {
        "data": data,
        "params": params,
    }}

    return _run_user_code(entry_point=entry_point, entry_point_arguments=entry_point_arguments)



import base64
import json
import os
from typing import Dict, Union

from loguru import logger


def any_to_json(data: any) -> str:
    return json.dumps(data, default=lambda o: getattr(o, "__dict__", str(o)), sort_keys=True, indent=2)


def json_to_dict(value: str, base64_encoded: bool) -> Dict:
    if value is None:
        return {}
    decoded_value = value
    if base64_encoded:
        try:
            decoded_value = base64.urlsafe_b64decode(value).decode("utf-8")
        except Exception as e:
            raise ValueError(f"Not a base64 string? - {e}")
    try:
        return json.loads(decoded_value)
    except Exception as e:
        raise ValueError(f"Not a json string? {e}")


class JsonEnvironmentReader:
    def __init__(self, name: str, base64_encoded: bool):
        self.__name = name
        self.__base64_encoded = base64_encoded

    def __enter__(self):
        return self

    def __exit__(self, resource_type, value, tb):
        pass

    def read(self) -> Union[Dict, None]:
        value = os.environ.get(self.__name, None)
        if value is None:
            return None
        try:
            return json_to_dict(value, self.__base64_encoded)
        except ValueError as e:
            logger.error("Error reading environment variable '{}': {}", self.__name, e)
            return None



import os
from abc import abstractmethod, ABC

from loguru import logger

from planqk.commons.runtime.file import FileReader
from planqk.commons.runtime.json import JsonEnvironmentReader


class InputFileReader(ABC):
    def __init__(self, file_path: str, base64_encoded: bool):
        self._file_path = file_path
        self._base64_encoded = base64_encoded

        logger.debug(f"Base64 encoded data? {base64_encoded}")

    def __exit__(self, resource_type, value, tb):
        pass

    def __enter__(self):
        return self

    @abstractmethod
    def read(self):
        pass


class InputDataReader(InputFileReader, ABC):
    def read(self):
        input_data = None

        if os.path.isfile(self._file_path):
            logger.info(f"Using input data from file '{self._file_path}'")
            with FileReader(["/var/input/data.json", "./input/data.json", self._file_path], self._base64_encoded) as reader:
                input_data = reader.read_to_dict()
        else:
            if "DATA_VALUE" in os.environ:
                logger.info("Using input data from environment variable 'DATA_VALUE'")
                with JsonEnvironmentReader("DATA_VALUE", self._base64_encoded) as reader:
                    input_data = reader.read()
            elif "INPUT_DATA" in os.environ:
                logger.warning("DEPRECATED: Using input data from environment variable 'INPUT_DATA'")
                with JsonEnvironmentReader("INPUT_DATA", self._base64_encoded) as reader:
                    input_data = reader.read()

        if input_data is None:
            logger.warning("No input data found, working with empty dict")
            return {}

        return input_data


class InputParamsReader(InputFileReader, ABC):
    def read(self):
        input_params = None

        if os.path.isfile(self._file_path):
            logger.info(f"Using input params from file '{self._file_path}'")
            with FileReader(["/var/input/params.json", "./input/params.json", self._file_path], self._base64_encoded) as reader:
                input_params = reader.read_to_dict()
        else:
            if "PARAMS_VALUE" in os.environ:
                logger.info("Using input params from environment variable 'PARAMS_VALUE'")
                with JsonEnvironmentReader("PARAMS_VALUE", self._base64_encoded) as reader:
                    input_params = reader.read()
            elif "INPUT_PARAMS" in os.environ:
                logger.warning("DEPRECATED: Using input params from environment variable 'INPUT_PARAMS'")
                with JsonEnvironmentReader("INPUT_PARAMS", self._base64_encoded) as reader:
                    input_params = reader.read()

        if input_params is None:
            logger.warning("No input parameters found, working with empty dict")
            return {}

        return input_params



import os
import traceback

from loguru import logger

from planqk.commons import __version__
from planqk.commons.runtime.input import InputDataReader, InputParamsReader
from planqk.commons.runtime.response import ResponseHandler
from planqk.commons.runtime.string import str_to_bool
from planqk.commons.runtime.user_code import run


def main():
    logger.debug(f"Template Version: {__version__}")

    entry_point = os.environ.get("ENTRY_POINT", "user_code.src.program:run")
    logger.debug(f"Entry Point: {entry_point}")

    data_file = os.environ.get("DATA_FILE", "/var/input/data.json")
    params_file = os.environ.get("PARAMS_FILE", "/var/input/params.json")
    base64_encoded = str_to_bool(os.environ.get("BASE64_ENCODED", "true"))
    data_base64_encoded = str_to_bool(os.environ.get("DATA_BASE64_ENCODED", "true"))
    params_base64_encoded = str_to_bool(os.environ.get("PARAMS_BASE64_ENCODED", "true"))

    with InputDataReader(data_file, data_base64_encoded and base64_encoded) as reader:
        input_data = reader.read()

    with InputParamsReader(params_file, params_base64_encoded and base64_encoded) as reader:
        input_params = reader.read()

    response = None
    try:
        response = run(entry_point, input_data, input_params)
    except Exception as e:
        logger.error(f"Error executing user code: {e}")
        traceback.print_exc()
        exit(1)

    response_handler = ResponseHandler(response)

    if not response_handler.is_response():
        logger.warning("Result type is not one of ResultResponse or ErrorResponse")

    response_handler.print_json()

    if response_handler.is_error_response():
        exit(1)



from planqk.commons.runtime.json import any_to_json


class Response:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def json(self):
        return any_to_json(self)


class ErrorResponse(Response):
    """
    Represents an error to be passed back to the caller.
    """
    pass


class ResultResponse(Response):
    """
    Represents the result to be passed back to the caller.
    """
    pass


class ResponseHandler:
    def __init__(self, result):
        self.__result = result
        self.__class_name = None
        if result is not None:
            self.__class_name = result.__class__.__name__

    def is_result_response(self) -> bool:
        return self.__class_name == "ResultResponse"

    def is_error_response(self) -> bool:
        return self.__class_name == "ErrorResponse"

    def is_response(self) -> bool:
        return self.is_result_response() or self.is_error_response()

    def print_json(self):
        print(str(self))

    def __str__(self):
        if self.__result is None:
            return ""
        else:
            return f"\nPlanQK:Job:MultilineResult\n{any_to_json(self.__result)}\nPlanQK:Job:MultilineResult"



import os

from planqk.commons.runtime import main, ResultResponse, init_logging


def run(**kwargs) -> any:
    return ResultResponse()


def test_main():
    # set the entry point to the test function
    os.environ["ENTRY_POINT"] = "tests.test_main:run"

    init_logging()

    print()
    main()

    # reset the entry point
    os.environ["ENTRY_POINT"] = "user_code.src.program:run"


if __name__ == "__main__":
    test_main()



from setuptools import find_namespace_packages, setup

with open("./README.md", "r") as fh:
    long_description = fh.read()

with open("./requirements.txt", "r") as fh:
    requirements = fh.readlines()

setup(
    name="planqk-commons",
    version="1.1.1",
    author="Kipu Quantum GmbH",
    author_email="info@kipu-quantum.com",
    url="https://gitlab.com/planqk-foss/planqk-commons",
    description="Set of common utilities and classes to boost development of quantum computing applications using the PLANQK Platform.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["planqk", "planqk.*"]),
    license="apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires=">=3.9",
)
